import _ from "lodash";

import { Chess } from "../chessops/src/chess";
import { Move } from "../chessops/src/types";

import {
  Atomic,
  Antichess,
  Crazyhouse,
  Horde,
  KingOfTheHill,
  RacingKings,
  ThreeCheck,
} from "../chessops/src/variant";

import { parseSan, makeSan } from "../chessops/src/san";

import { makeFen, parseFen } from "../chessops/src/fen";

import { parseUci, makeUci } from "../chessops/src/util";

import { chessgroundDests } from "../chessops/src/compat";

export { chessgroundDests };

export type Uci = string;
export type San = string;
export type Variant = string;
export type Fen = string;
export type GameNodeId = string;
export type Shape = string;

const PROMOTION_PIECES = ["q", "r", "b", "n", "k"];
const PROMOTION_PIECES_EXT = [""].concat(PROMOTION_PIECES);

// Pos_ is an abstraction of a chess position
export class Pos_ {
  pos: Chess;

  constructor() {
    // initialize to standard chess starting position
    this.pos = Chess.default();
  }

  get rules(): Variant {
    return this.pos.rules;
  }

  get variant(): Variant {
    return this.rules;
  }

  rawLegalUcis() {
    return Array.from(chessgroundDests(this.pos).entries())
      .map((entry) => entry[1].map((dest) => `${entry[0]}${dest}`))
      .flat();
  }

  setVariant(variant: Variant) {
    switch (variant) {
      case "atomic":
        this.pos = Atomic.default();
        break;
      case "antichess":
        this.pos = Antichess.default();
        break;
      case "crazyhouse":
        this.pos = Crazyhouse.default();
        break;
      case "horde":
        this.pos = Horde.default();
        break;
      case "kingofthehill":
        this.pos = KingOfTheHill.default();
        break;
      case "racingkings":
        this.pos = RacingKings.default();
        break;
      case "3check":
      case "threecheck":
        this.pos = ThreeCheck.default();
        break;
      default:
        this.pos = Chess.default();
    }
    return this;
  }

  setFen(fen: Fen) {
    const setupResult = parseFen(fen);

    if (setupResult.isErr) {
      return this;
    }

    const setup = setupResult.value;

    let posResult = undefined;

    const variant = this.pos.rules;

    switch (variant) {
      case "atomic":
        posResult = Atomic.fromSetup(setup);
        break;
      case "antichess":
        posResult = Antichess.fromSetup(setup);
        break;
      case "crazyhouse":
        posResult = Crazyhouse.fromSetup(setup);
        break;
      case "horde":
        posResult = Horde.fromSetup(setup);
        break;
      case "kingofthehill":
        posResult = KingOfTheHill.fromSetup(setup);
        break;
      case "racingkings":
        posResult = RacingKings.fromSetup(setup);
        break;
      case "3check":
        posResult = ThreeCheck.fromSetup(setup);
        break;
      default:
        posResult = Chess.fromSetup(setup);
    }

    if (posResult === undefined) return this;

    if (posResult.isErr) return this;

    this.pos = posResult.value;

    return this;
  }

  reportFen() {
    return makeFen(this.pos.toSetup());
  }

  sanToMove(san: San): Move | undefined {
    return parseSan(this.pos, san);
  }

  moveToSan(move: Move) {
    return makeSan(this.pos, move);
  }

  uciToMove(uci: Uci): Move | undefined {
    const move = parseUci(uci);
    return move;
  }

  moveToUci(move: Move) {
    return makeUci(move);
  }

  play(move: Move) {
    this.pos.play(move);
    return this;
  }

  playSan(san: San) {
    const move = this.sanToMove(san);
    if (move === undefined) return this;
    return this.play(move);
  }

  playUci(uci: Uci) {
    const move = this.uciToMove(uci);
    if (move === undefined) return this;
    return this.play(move);
  }

  sanToUci(san: San): Uci | undefined {
    const move = this.sanToMove(san);
    if (move === undefined) return undefined;
    return this.moveToUci(move);
  }

  uciToSan(uci: Uci): San | undefined {
    const move = this.uciToMove(uci);
    if (move === undefined) return undefined;
    return this.moveToSan(move);
  }

  toString() {
    return `[Pos ${this.pos.rules} ${this.reportFen()}]`;
  }

  legalsForUci(uci: Uci) {
    return PROMOTION_PIECES_EXT.filter((pp) => {
      const move = this.uciToMove(`${uci}${pp}`);
      if (move === undefined) return false;
      const legal = this.pos.isLegal(move, this.pos.ctx());
      return legal;
    }).map((pp) => `${uci}${pp}`);
  }

  isUciLegal(uci: Uci): boolean {
    return this.legalsForUci(uci).length > 0;
  }

  isSanLegal(san: San): boolean {
    const uci = this.sanToUci(san);
    if (uci === undefined) return false;
    return this.isUciLegal(uci);
  }

  allLegalUcis() {
    return this.rawLegalUcis()
      .map((uci) => this.legalsForUci(uci))
      .flat();
  }

  allLegalSans() {
    const rawLegalSans = this.allLegalUcis().map((uci) => this.uciToSan(uci));
    const legalSans = _.uniq(rawLegalSans);
    return legalSans;
  }
}

export function Pos() {
  return new Pos_();
}

export class Game_ {
  variant: Variant;
  pos: Pos_;
  root: GameNode_;
  current: GameNode_;
  nodes: { [id: string]: GameNode_ };

  constructor() {
    this.fromProps({});
  }

  fromProps(props: any) {
    this.nodes = {};

    this.variant = props.variant || "chess";

    this.pos = Pos().setVariant(this.variant);
    if (props.fen) this.pos.setFen(props.fen);

    if (props.nodes) {
      for (const id in props.nodes) {
        this.nodes[id] = GameNode(this).fromProps(props.nodes[id]);
      }
    }

    if (props.root) {
      this.root = GameNode(this).fromProps(props.root);
    } else {
      this.root = GameNode(this).fromProps({
        fen: this.pos.reportFen(),
        id: "*",
      });
    }

    if (props.current) {
      this.current = GameNode(this).fromProps(props.current);
    } else {
      this.current = this.root;
    }

    this.nodes[this.root.id] = this.root;

    this.pos.setFen(this.current.fen);

    return this;
  }

  serialize(): any {
    const nodesSerialized: any = {};
    for (const id in this.nodes) {
      nodesSerialized[id] = this.nodes[id].serialize();
    }
    return {
      variant: this.variant,
      root: this.root.serialize(),
      current: this.current.serialize(),
      nodes: nodesSerialized,
    };
  }

  stringify(): string {
    return JSON.stringify(this.serialize(), null, 2);
  }

  parse(json: string): Game_ {
    if (!json) {
      return this.fromProps({});
    }
    const blob = JSON.parse(json);
    this.fromProps(blob);
    return this;
  }

  reportFen(): Fen {
    return this.pos.reportFen();
  }

  prev(): GameNode_ {
    return this.current.prev();
  }

  next(): GameNode_ {
    return this.current.next();
  }

  setCurrent(node: GameNode_): Game_ {
    this.current = node;
    this.pos.setFen(this.current.fen);
    return this;
  }

  selectId(id: GameNodeId): Game_ {
    if (id in this.nodes) {
      this.current = this.nodes[id];
    }
    return this.setCurrent(this.current);
  }

  back(): boolean {
    const id = this.current.id;
    this.setCurrent(this.prev());
    return id !== this.current.id;
  }

  toBegin(): boolean {
    let ok = false;
    while (this.back()) {
      ok = true;
    }
    return ok;
  }

  forward(): boolean {
    const id = this.current.id;
    this.setCurrent(this.next());
    return id !== this.current.id;
  }

  toEnd(): boolean {
    let ok = false;
    while (this.forward()) {
      ok = true;
    }
    return ok;
  }

  del(): boolean {
    const oldId = this.current.id;
    if (!this.back()) return false;
    const len = oldId.length;
    for (const id in this.nodes) {
      if (id.substring(0, len) === oldId) {
        delete this.nodes[id];
      }
    }
    this.current.childs = [];
    return true;
  }

  mainLine(): GameNode_[] {
    let ptr = this.root;
    let line = [];
    do {
      line.push(ptr);
      if (ptr.isTerminal) {
        return line;
      }
      ptr = ptr.next();
    } while (true);
  }

  setVariant(variant: Variant, fen: Fen | undefined): Game_ {
    this.fromProps({
      variant,
      fen,
    });
    return this;
  }

  playSans(sans: San[]): Game_ {
    sans.forEach((san) => this.playSan(san));
    return this;
  }

  playSansStr(sans: string): Game_ {
    this.playSans(sans.split(" "));
    return this;
  }

  playSan(san: San): boolean {
    if (!this.pos.isSanLegal(san)) return false;
    const uci = this.pos.sanToUci(san);
    this.pos.playSan(san);
    const newId = this.current.id + "_" + san;
    const oldCurrent = this.current;
    if (this.nodes[newId]) {
      this.current = this.nodes[newId];
    } else {
      this.current = GameNode(this).fromProps({
        fen: this.reportFen(),
        genUci: uci,
        genSan: san,
        id: newId,
      });
      this.nodes[newId] = this.current;
    }
    oldCurrent.bringForwardChild(newId);
    return true;
  }

  playUci(uci: Uci): boolean {
    if (!this.pos.isUciLegal(uci)) return false;
    const san = this.pos.uciToSan(uci);
    if (san === undefined) return false;
    return this.playSan(san);
  }

  clone() {
    return Game().parse(this.stringify());
  }
}

export function Game() {
  return new Game_();
}

export class GameNode_ {
  id: GameNodeId;
  fen: Fen;
  genUci: Uci;
  genSan: San;
  weight: number;
  priority: number;
  comment: string;
  shapes: Shape[];
  childs: GameNodeId[];

  parentGame: Game_;

  constructor(parentGame: Game_) {
    this.parentGame = parentGame;
    this.fromProps({});
  }

  clone() {
    return GameNode(this.parentGame).parse(this.stringify());
  }

  fromProps(props: any) {
    this.id = props.id;
    this.fen = props.fen;
    this.genUci = props.genUci;
    this.genSan = props.genSan;
    this.weight = typeof props.weight === "number" ? props.weight : 0;
    this.priority = typeof props.priority === "number" ? props.priority : 0;
    this.comment = props.comment || "";
    this.shapes = props.shapes || [];
    this.childs = props.childs || [];
    return this;
  }

  serialize(): any {
    return {
      id: this.id,
      fen: this.fen,
      genUci: this.genUci,
      genSan: this.genSan,
      weight: this.weight,
      priority: this.priority,
      comment: this.comment,
      shapes: this.shapes,
      childs: this.childs,
    };
  }

  stringify(): string {
    return JSON.stringify(this.serialize(), null, 2);
  }

  parse(json: string): GameNode_ {
    if (!json) {
      return this.fromProps({});
    }
    const blob: any = JSON.parse(json);
    return this.fromProps(blob);
  }

  prevId(): GameNodeId {
    if (this.id === "*") return this.id;
    const sans = this.id.split("_");
    sans.pop();
    return sans.join("_");
  }

  addChild(childId: GameNodeId): GameNode_ {
    if (this.childs.find((testId) => testId === childId)) return this;
    this.childs.push(childId);
    return this;
  }

  removeChild(childId: GameNodeId): GameNode_ {
    if (this.childs.find((testId) => testId === childId)) {
      this.childs = this.childs.filter((child) => child !== childId);
    }
    return this;
  }

  bringForwardChild(childId: GameNodeId): GameNode_ {
    this.removeChild(childId);
    this.childs.unshift(childId);
    return this;
  }

  get isTerminal(): boolean {
    return this.childs.length === 0;
  }

  next(): GameNode_ {
    if (this.isTerminal) return this;
    return this.nexts()[0];
  }

  get isRoot(): boolean {
    return this.id === "*";
  }

  prev(): GameNode_ {
    return this.parentGame.nodes[this.prevId()];
  }

  nexts(): GameNode_[] {
    if (this.isTerminal) return [];
    return [this.parentGame.nodes[this.childs[0]]];
  }

  nextOrphans(): GameNode_[] {
    if (this.childs.length < 2) return [];
    return this.childs.slice(1).map((id) => this.parentGame.nodes[id]);
  }

  get displaySan(): San {
    if (this.isRoot) return "*";
    return this.genSan;
  }
}

export function GameNode(parentGame: Game_) {
  return new GameNode_(parentGame);
}
